/**
 * COMPLETE MIN SAT SOLVER
 * 
 * Author: Gabriel Corvera
 * Cornell University / ITESM CEM
 * 
 */
  #include <math.h>
  #include <stdio.h>
  #include <errno.h>
  #include <stdlib.h>
  #include <string.h>
  #include "gscminsat.h"
  
  /**
   * readCNF. This function returns true if the file was successfully loaded into the data structures
   */
  boolean readCNF(char *filename){
 	FILE *f;	
 	int i,j,k,temp,t_array[MAX_CLAUSE_VARS];
 	char c,s[200];
 	clause *temporalClause;
 	/*if the file did not open I should return false*/
 	if ((f = fopen (filename, "r")) == NULL) {
    	printf("\nc readCNF - ERROR OPENING FILE '%s'", filename); 
    	return false; 
    }
    /*Begin reading comments in the input file*/
 	fscanf(f, "%c", &c);
  	while(c == 'c') {
    	do
      		fscanf(f, "%c", &c);
    	while(c != '\n');
    	fscanf(f, "%c", &c);
  	}    
  	/*if we are not reading comments we should be in the p line*/
  	if(c != 'p') { 
  		printf("\nc readCNF - ERROR OPENING FILE '%s'", filename); 
  		return false; 
  	}
  	fseek(f, (long) (-1), SEEK_CUR);
  	fgets(s, 6, f);
  	/*we verify if the file is in cnf format.*/
  	if(strcmp("p cnf",s) || (feof(f))) { 
  		printf("\nc readCNF - BAD FILE FORMAT '%s'", filename); 
  		return false; 
  	}
  	/*The file shouldn't finish here*/
  	if(feof(f)) {
    	printf("\nc readCNF - BAD FILE FORMAT '%s'", filename);
   		return false;
  	}
  	fscanf(f, "%d%d", &numVars, &numClauses);
  	
  	/* assing enough space for an int array for the variables*/
  	variables = (int *)malloc(sizeof(int)*(numVars+1));
  	minsat = (int *)malloc(sizeof(int)*(numVars+1));
  	/* initialize with ceros */
 	for(i=0;i<numVars+1;i++){
		variables[i]=0;
		minsat[i]=0;
 	}  	
  	
  	/*time to read the clauses and fill the datastructures*/
  	for(i=0;i<numClauses;i++){
		if(feof(f)){
			return false;
		}
		j = 0;
		/*prepare space to store clause*/
		temporalClause = (clause *)malloc(sizeof(clause));
		temporalClause->id=i+1;
		do{
		/*read the integer of the formula*/
	      	fscanf(f, "%d", &temp);
	      	t_array[j++]=temp;
		}while(temp!=0);
		t_array[j]=0;
		for(k=0;k<j;k++){
			temporalClause->vars[k]=t_array[k];
		}
		temporalClause->varsSize=j-1;
		/*assign the next clause*/
		temporalClause->next=clauses;
		clauses=temporalClause;
		
  	}
  	return true;		
 }

 int findNumberOfSATClauses(){
 	int i,sat_clauses=0,var;
 	clause *cl_pointer; 	
 	cl_pointer=clauses;
 	
 	/*search every clause to see how many clauses are satisfied by the actual 
 	 variable assignation*/
 	while(cl_pointer!=NULL){
	 	for(i=0;i<cl_pointer->varsSize;i++){
	 		var = cl_pointer->vars[i];
 			if(var>0){

 				/*the variable var is true! then the clause is sat 	*/			
 				if(variables[var]){
 					sat_clauses++;
 					break;
 				} 				
 			}else if(var<0){
				/*the variable var is false! then the clause is sat */
 				if(!variables[-var]){
 					sat_clauses++;
 					break;
 				} 				 				
 			}
 			
 		}
 		cl_pointer=cl_pointer->next;
 	}	
 	return sat_clauses;
 }
 

 /* Change the actual truth assignment shifting 
  * 1 bit in the array like adding 1 to a binary 
  * number 
  */
 void changeTruthAssignment(){
 	int i;
 	for(i=1;i<numVars+1;i++){
 		if(variables[i]==0){
 			variables[i]=1;	
 			break;
 		}
 		else{
 			variables[i]=0;			
 		}	 		
 	} 	
 }

 void printAssignmentsAndSATClauses(){
 	long i,j;
 	for(i=0;i<pow(2,numVars);i++){
 		changeTruthAssignment();	
 		printf("c Truth Assignment: ");
 		for(j=1;j<numVars+1;j++){
 			printf("%i ",variables[j]);
 		}
 		printf("\n");
		printf("c Number of SAT Clauses for this truth assignment: %i\n",findNumberOfSATClauses()); 		
 	}	
 }
 
 void copyAssignmentToSolution(){
 	int i;
 	for(i=1;i<numVars+1;i++){
 		minsat[i]=variables[i];
 	} 		
 }
 
 int findMINSAT(){
 	long i,sat_clauses,min=numClauses;
 	for(i=0;i<pow(2,numVars);i++){
 		changeTruthAssignment();	
		sat_clauses = findNumberOfSATClauses();
		if(sat_clauses<min){
			min = sat_clauses;
			/* we found a possible minsat assignment, copy
			 * the actual solution to the solution array
			 * to keep it for future reference
			 */
			copyAssignmentToSolution();
		}
 	}	 	
 	return min;
 }

 void printSolution(){
 	int i;
 	for(i=1;i<numVars+1;i++){
 		printf("%d ",minsat[i]);
 	}
 	printf("\n");
 		
 }

 int main (int argc,char * args[]){
 	int j;
 	if(argc!=2){
 		printf("USAGE: %s filename.cnf\n",args[0]); 
 		exit(0);
 	}
 	if(!readCNF(args[1])){
 		return false;
 	}	 	 	
 	#ifdef DEBUG
	 	printAssignmentsAndSATClauses();
	#endif
	printf("c The MinSAT value of this file is: %d\n",findMINSAT());
	printf("c The Truth Assignment found to be minsat is: ");
	printSolution();	
 	return true;
 }
